"""Base classes for searchable element implementations"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, List, Tuple, Union

from selenium.webdriver import Remote
from selenium.webdriver.remote.webelement import WebElement

Wrapped = Union[WebElement, List[WebElement], Remote]
BROWSER = "Browser"


class Searchable(ABC):
    """Base class for objects that presents lazy-loading searchable elements"""

    __cached__: Wrapped = None
    __is_browser__ = False
    _locator: LocatorStrategy  # pylint: disable=used-before-assignment

    def __init__(self, locator: LocatorStrategy):
        self._locator = locator

    def get_actual(self):
        """Get actual instance of wrapped class"""
        if self.__cached__ is None:
            self.__cached__ = self._search()
        return self.__cached__

    def _search(self) -> Wrapped:
        return self._locator.get()

    @property
    def browser(self):
        """Return used browser instance"""
        # pylint: disable=protected-access
        parent_locator = self._locator
        while not parent_locator.context.__is_browser__:
            parent_locator = parent_locator.context._locator
        return parent_locator.context


class LocatorStrategy(ABC):
    """Base class for locator containers"""

    context: Searchable

    def __init__(self, by: Tuple[str, str], context: Searchable):
        self.context = context
        self.by = by

    @abstractmethod
    def get(self) -> Any:
        """"Run locator and return object"""

    @abstractmethod
    def __repr__(self) -> str: ...  # pylint: disable=multiple-statements



"""Lazy locator wrappers"""
from abc import ABC, abstractmethod
from typing import Any, Callable, List, Union
from weakref import proxy

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.remote.webelement import WebElement

from pyasli.elements.searchable import LocatorStrategy, Searchable

# pylint: disable=protected-access


def _search_in_context(context: Searchable, method: str, by: tuple) -> List[WebElement]:
    actual = context.get_actual()
    if not isinstance(actual, list):
        actual: List[WebElement] = [actual]
    result = []
    for elem in actual:
        try:
            found = getattr(elem, method)(*by)
        except NoSuchElementException:
            continue
        if isinstance(found, list):
            result.extend(found)  # append or extend depending on result of search
        else:
            result.append(found)
    if not result:
        raise NoSuchElementException
    return result


class SingleElementLocator(LocatorStrategy):
    """Locator for returning single element"""

    def get(self) -> WebElement:
        """Get single matching web element"""
        result = _search_in_context(self.context, "find_element", self.by)
        return result[0]

    def __repr__(self):
        # recursive repr
        return f"{repr(self.context._locator)} -> {self.by}"


class MultipleElementLocator(LocatorStrategy):
    """Locator strategy for multiple elements"""

    def get(self) -> List[WebElement]:
        """Get list of matching web elements"""
        return _search_in_context(self.context, "find_elements", self.by)

    def __repr__(self):
        # recursive repr
        return f"{repr(self.context._locator)} -> [{self.by}]"


# SUB ELEMENTS

class SubElementLocator(LocatorStrategy):
    """Locator for elements extracted from element list"""

    def __init__(self, whole: MultipleElementLocator, sub: Union[slice, int]):
        super().__init__(whole.by, whole.context)
        self._whole = whole
        self._sub = sub

    def get(self) -> List[WebElement]:
        """Get slice from element collection (not going deeper)"""
        elements = self._whole.get()
        return elements[self._sub]

    def __repr__(self):
        # recursive repr
        return f"{repr(self._whole)}[{self._sub}]"


class SlicedElementLocator(SubElementLocator, MultipleElementLocator):
    """Locator for returning slice from element collection"""

    def __init__(self, whole: MultipleElementLocator, slize: slice):
        super().__init__(whole, slize)

    def __repr__(self):
        # recursive repr
        start = "" if self._sub.start is None else self._sub.start
        stop = "" if self._sub.stop is None else self._sub.stop
        sss = f"{start}:{stop}"
        if self._sub.step:
            sss = f"{sss}:{self._sub.step}"
        return f"{repr(self._whole)}[{sss}]"


class IndexElementLocator(SubElementLocator, SingleElementLocator):
    """Locator for returning element by index from collection"""

    def __init__(self, whole: MultipleElementLocator, index: int):
        super().__init__(whole, index)


# CONDITIONS


class ConditionLocator(LocatorStrategy, ABC):
    """Filtering and finding element of collection locator"""
    _condition: Callable[[Any], bool]

    def __init__(self, collection, condition: Callable[[Any], bool]):
        self._whole = proxy(collection._locator)
        super().__init__(self._whole.by, self._whole.context)
        self._collection = collection
        self._condition = condition

    @property
    @abstractmethod
    def _word(self) -> str: ...  # pylint:disable=multiple-statements

    def __repr__(self):
        return f"{self._whole.__repr__()}.{self._word}({self._condition.__name__})"

    def get(self) -> List[WebElement]:
        """Get only web elements matching condition"""
        raise NotImplementedError


class FilteredCollectionLocator(ConditionLocator, MultipleElementLocator):
    """Locator for returning only elements of collection matching condition"""

    @property
    def _word(self):
        return "filter"

    def get(self) -> List[WebElement]:
        """Get only web elements matching condition"""
        filtered = [elem.get_actual() for elem in self._collection if self._condition(elem)]
        return filtered


class FindElementLocator(ConditionLocator, SingleElementLocator):
    """Locator for returning single element of collection matching condition"""

    @property
    def _word(self):
        return "find"

    def get(self) -> WebElement:
        """Get only single element matching condition"""
        for elem in self._collection:
            if self._condition(elem):
                return elem.get_actual()
        raise NoSuchElementException(f"No element matching {repr(self)}")



import abc
import logging
import os
import uuid
import warnings

import wrapt
from selenium.common.exceptions import WebDriverException

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.ERROR)
__STH = logging.StreamHandler()
__STH.setLevel(logging.ERROR)
LOGGER.addHandler(__STH)


class NoBrowserException(WebDriverException):
    """No operable browser is open"""


def _save_screenshot(data: bytes) -> str:
    os.makedirs("./logs", exist_ok=True)
    path = os.path.abspath(f"./logs/{uuid.uuid4()}.png")
    with open(path, "wb+") as out:
        out.write(data)
    return path


class Screenshotable(abc.ABC):
    """Object instance provide screenshot functionality"""

    @abc.abstractmethod
    def capture_screenshot(self) -> bytes: ...


@wrapt.decorator
def screenshot_on_fail(wrapped, instance: Screenshotable = None, args=None, kwargs=None):
    """Capture screenshot on method error

    One of the following errors handled: WebDriverException, AssertionError, TimeoutError
    """
    try:
        return wrapped(*args, **kwargs)
    except (WebDriverException, AssertionError, TimeoutError) as sc_e:
        if not hasattr(instance, 'capture_screenshot'):
            warnings.warn(f'`capture_screenshot` method is missing for {instance}.'
                          'No screenshot can be captured')
            raise sc_e
        screenshot = instance.capture_screenshot()
        LOGGER.exception("Screenshot captured on failure: %s", _save_screenshot(screenshot))
        raise sc_e



"""Shared browser, safe to use with pytest-xdist"""
from .browser_session import BrowserSession

browser = BrowserSession()  # pylint: disable=invalid-name


class _BrowserState:
    """Static container for browser state"""

    # only browser_name, lol
    browser_name = "chrome"  # please, don't change me directly

    def __new__(cls):
        raise NotImplementedError


def set_browser(browser_name: str):
    """Set shared browser. More precise configuration can be reached with `browser.setup_browser`"""
    if browser_name != _BrowserState.browser_name:
        browser.close_all_windows()
        _BrowserState.browser_name = browser_name
    browser.setup_browser(browser_name)



"""Browser management"""

from __future__ import annotations

import atexit
import re
from contextlib import AbstractContextManager
from typing import Any, Dict, NamedTuple, Optional, Type, Union

from selenium.webdriver import (
    Chrome, ChromeOptions, DesiredCapabilities, Firefox, FirefoxOptions, Ie, IeOptions, Remote
)
from selenium.webdriver.remote.webdriver import WebDriver
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import IEDriverManager

from pyasli.bys import CssSelectorOrBy
from pyasli.elements.elements import Element, ElementCollection, FindElementsMixin
from pyasli.elements.searchable import LocatorStrategy, Searchable
from pyasli.exceptions import NoBrowserException

_BROWSER_MAPPING: Dict[str, Type[Remote]] = {
    "chrome": Chrome,
    "ie": Ie,
    "firefox": Firefox,
    "remote": Remote,
}

_OPTIONS_MAPPING = {
    "chrome": ChromeOptions,
    "ie": IeOptions,
    "firefox": FirefoxOptions,
}

_MANAGER_MAPPING = {
    "chrome": ChromeDriverManager,
    "ie": IEDriverManager,
    "firefox": GeckoDriverManager,
}

_FULL_URL_RE = re.compile(r"http(s)?://.+")


# noinspection PyTypeChecker
class BrowserLocatorStrategy(LocatorStrategy):
    """Root level locator strategy"""
    context: BrowserSession  # pylint: disable=used-before-assignment

    def get(self) -> Any:
        """Return actual browser"""
        raise NotImplementedError  # should not be used

    def __repr__(self) -> str:
        return self.context.browser_name.capitalize()

    def __init__(self, browser_session: BrowserSession):
        super().__init__(None, browser_session)


class BrowserSession(Searchable, FindElementsMixin, AbstractContextManager):
    """Lazy webdriver wrapper"""

    _actual: Optional[Remote] = None
    __is_browser__ = True
    browser_name: str = None
    options = None
    desired_capabilities = None
    _other_options: dict = None

    base_url: str = None

    @property
    def browser(self) -> BrowserSession:
        """Browser of browser is self XD"""
        return self  # pragma: no cover

    def _check_running(self):
        if self._actual is None:
            raise NoBrowserException("You should open some page before doing anything")

    def __init__(self, browser="chrome", base_url=None):
        """Init new lazy browser session.
        Setup browser to be used to given local headless browser
        """
        super().__init__(BrowserLocatorStrategy(self))
        self.setup_browser(browser)
        self.base_url = base_url
        atexit.register(self.close_all_windows)

    def setup_browser(self, browser: str, remote=False, headless=True,
                      options: Union[ChromeOptions, FirefoxOptions, IeOptions] = None,
                      desired_capabilities: DesiredCapabilities = None,
                      **other_options):
        """Configure browser to be used

        :param str browser: Name of browser to be used
        :param bool remote: Is used browser running on remote host.
            In this case you should set `command_executor` argument to desired host value
        :param bool headless: If `options` are not set, control `headless` option
        :param options: Browser options
        :param desired_capabilities: Browser desired capabilities
        :param other_options: Other options which will be passed to WebDriver constructor
        """
        if remote:
            self.browser_name = "remote"
            self.options = options or _OPTIONS_MAPPING[browser]()
        else:
            self.browser_name = browser
            self.options = options
        if options is None:
            self.options = _OPTIONS_MAPPING[browser]()
            self.options.headless = headless
        self.desired_capabilities = desired_capabilities
        self._other_options = other_options

    def __init_browser(self):
        """Start new browser instance"""
        if self._actual is not None:
            return
        browser_cls = _BROWSER_MAPPING[self.browser_name]
        if browser_cls is Remote:
            self._actual = Remote(desired_capabilities=self.desired_capabilities, options=self.options,
                                  **self._other_options)
            return
        driver_path = _MANAGER_MAPPING[self.browser_name]().install()
        if browser_cls is Firefox:
            self._actual = Firefox(executable_path=driver_path, options=self.options,
                                   desired_capabilities=self.desired_capabilities, **self._other_options)
        else:
            self._actual = browser_cls(driver_path, options=self.options, **self._other_options)

    def set_driver(self, webdriver: Remote):
        """Override lazy driver initialization with already initialized webdriver"""
        if self._actual is not None:
            self._actual.quit()
        self._actual = webdriver

    def open(self, url: str):
        """Open given URL"""
        self.__init_browser()

        if self.base_url and not _FULL_URL_RE.fullmatch(url):
            if not url.startswith("/"):
                url = f"/{url}"
            url = f"{self.base_url}{url}"
        self._actual.get(url)

    def element(self, by: CssSelectorOrBy) -> Element:
        """Find single element by locator (css selector by default)"""
        return super().element(by)

    def elements(self, by: CssSelectorOrBy) -> ElementCollection:
        """Find single element by locator (css selector by default)"""
        return super().elements(by)

    def add_cookie(self, cookie_dict: dict):
        """Add cookies to cookie storage"""
        self._check_running()
        self._actual.add_cookie(cookie_dict)

    def close_window(self):
        """Close current browser window"""
        self._check_running()
        self._actual.close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close_all_windows()

    def close_all_windows(self):
        """Close all browser windows"""
        actual = self._actual
        if actual is not None:
            actual.quit()
            self._actual = None

    @property
    def url(self) -> URL:
        """Get current page URL"""
        return URL(self._actual.current_url, self.base_url)

    def get_screenshot_as_png(self) -> bytes:
        """Capture whole browser page screenshot"""
        return self._actual.get_screenshot_as_png()

    capture_screenshot = get_screenshot_as_png

    def get_actual(self) -> WebDriver:
        """Get browser instance"""
        return self._actual


def _url_with_base(base_url: str, uri: str) -> str:
    if base_url is not None:
        if _FULL_URL_RE.fullmatch(uri):
            return uri
        if uri.startswith("/"):
            uri = uri[1:]
        return f"{base_url}/{uri}"
    return uri


class URL(NamedTuple):
    """Two-part URL"""
    url: str
    base_url: str = None

    def __str__(self):
        return _url_with_base(self.base_url, self.url)

    def __eq__(self, other):
        return str(self) == _url_with_base(self.base_url, str(other))

    def __ne__(self, other):
        return str(self) != _url_with_base(self.base_url, str(other))



"""List of helpers with commonly used conditions"""

from pyasli.elements.elements import Element, ElementCondition


# pylint: disable=invalid-name

def __visible(element: Element) -> bool:
    """Element is visible"""
    return element.visible


def __hidden(element: Element) -> bool:
    """Element is hidden"""
    return element.hidden


def __exists(element: Element) -> bool:
    """Element exists"""
    return element.exists


def __enabled(element: Element):
    """Condition to check if element is not disabled"""
    return element.enabled


def __disabled(element: Element):
    """Condition to check if element is disabled"""
    return element.disabled


def __missing(element: Element):
    """Condition to check if element is missing from DOM

    Antonym to `exists`
    """
    return not element.exists


def __clickable(element: Element) -> bool:
    """Check if element available for interaction"""
    return __visible(element) and __enabled(element)


def __rename(fnc, name):
    fnc.__name__ = name
    return fnc


# looks stupid, but this way PyCharm won't add brackets automatically
visible = __rename(__visible, "visible")
hidden = __rename(__hidden, "hidden")
exist = __rename(__exists, "exist")
missing = __rename(__missing, "missing")
enabled = __rename(__enabled, "enabled")
disabled = __rename(__disabled, "disabled")
clickable = __rename(__clickable, "clickable")


def text_is(text: str) -> ElementCondition:
    """Element text is"""

    def _text_is(element):
        return element.text == text

    _text_is.__name__ = f"text_is '{text}'"  # condition name is used in repr
    return _text_is


def have_text(text: str) -> ElementCondition:
    """Contains text"""

    def _has_text(element):
        return text in element.text

    _has_text.__name__ = f"has_text '{text}'"
    return _has_text


has_text = have_text  # backward compatibility



"""Condition waiting"""
import time
from typing import Callable, TypeVar

T = TypeVar("T")


def wait_for(element: T, condition: Callable[[T], bool], timeout=5, exception=None):
    """Wait until condition for element is satisfied"""
    end_time = time.time() + timeout
    polling_time = 0.05
    while time.time() < end_time:
        if condition(element):
            return
        time.sleep(polling_time)
    message = f"Wait time has expired for condition `{condition.__name__}`"
    raise exception or TimeoutError(message)



"""Locator techniques"""
from typing import Tuple, Union

from selenium.webdriver.common.by import By

ByLocator = Tuple[str, str]

CssSelectorOrBy = Union[str, ByLocator]


def by_id(element_id: str) -> ByLocator:
    """By ID"""
    return By.ID, element_id


def by_css(css_selector: str) -> ByLocator:
    """By css selector"""
    return By.CSS_SELECTOR, css_selector


def by_xpath(xpath: str) -> ByLocator:
    """By xpath"""
    return By.XPATH, xpath


def by_name(name: str) -> ByLocator:
    """By name"""
    return By.NAME, name


def by_class(cls: str) -> ByLocator:
    """By class name"""
    return By.CLASS_NAME, cls


def by_tag(tag: str) -> ByLocator:
    """By tag name"""
    return By.TAG_NAME, tag



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyasli', 'pyasli.browsers', 'pyasli.elements']

package_data = \
{'': ['*']}

install_requires = \
['selenium>=3.141,<4.0', 'webdriver-manager>=1.7,<2.0', 'wrapt>=1.11,<2.0']

setup_kwargs = {
    'name': 'pyasli',
    'version': '0.2.5.1',
    'description': '(Python) Yet Another Selenium Instruments',
    'long_description': '## Pyasli\n[![coverage report](https://gitlab.com/outcatcher/pyasli/badges/master/coverage.svg)](https://outcatcher.gitlab.io/pyasli/)\n[![pipeline status](https://gitlab.com/outcatcher/pyasli/badges/master/pipeline.svg)](https://gitlab.com/outcatcher/pyasli/commits/master)\n[![PyPI version](https://img.shields.io/pypi/v/pyasli.svg)](https://pypi.org/project/pyasli/)\n\n### Simple selenium python wrapper\n\n#### There are two ways to use browser:\n\n##### Use default shared driver:\n\n```python\nfrom pyasli.browsers import browser\n\nbrowser.base_url = "https://the-internet.herokuapp.com"\nbrowser.open("/disappearing_elements")\nelement1 = browser.element("div.example p")\nassert element1.get_actual() is element1.get_actual(), "Element is found 2 times"\n```\n\n##### Use exact driver (can be used as context manager):\n```python\nfrom tests.conftest import browser_instance\n\nwith browser_instance(base_url="https://the-internet.herokuapp.com") as browser:\n    browser.open("/disappearing_elements")\n    element1 = browser.element("div.example p")\n    assert element1.get_actual() is element1.get_actual(), "Element is found 2 times"\n```\n\nIn case `browser_instance` is used as context manager, all browser windows will be closed at\nexiting context\n\n----\n\n##### There is no documentation currently. For usage please refer to tests\n\n----\n\nBuilt wheels are available at https://pypi.outcatcher.com/simple/pyasli/\n\n_<div>Icons made by <a href="https://www.freepik.com/" title="Freepik">Freepik</a> from <a href="https://www.flaticon.com/" \t\t\t    title="Flaticon">www.flaticon.com</a> is licensed by <a href="http://creativecommons.org/licenses/by/3.0/" \t\t\t    title="Creative Commons BY 3.0" target="_blank">CC 3.0 BY</a></div>_\n',
    'author': 'Anton Kachurin',
    'author_email': 'katchuring@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""WebElement wrappers"""
from __future__ import annotations

import time
from typing import Callable, Sequence, Union

import wrapt
from selenium.common.exceptions import NoSuchElementException, WebDriverException
from selenium.webdriver import ActionChains
from selenium.webdriver.remote.webelement import WebElement

from pyasli.bys import ByLocator, CssSelectorOrBy, by_css, by_xpath
from pyasli.elements.locators import (
    FilteredCollectionLocator, FindElementLocator, IndexElementLocator, MultipleElementLocator,
    SingleElementLocator, SlicedElementLocator
)
from pyasli.elements.searchable import Searchable
from pyasli.exceptions import NoBrowserException, Screenshotable, screenshot_on_fail
from pyasli.wait import wait_for


# pylint: disable=fixme

@wrapt.decorator
def _should_exist(wrapped, instance: Element = None, args=(), kwargs=None):
    """Check for element existence before run method

    This decorator uses `assure`, so screenshot will be taken automatically on fail
    """
    if instance is None:
        instance = args[0]

    def _exists(_el):
        return _el.exists

    instance.assure(_exists)
    return wrapped(*args, **kwargs)


def _css_to_by(by: CssSelectorOrBy) -> ByLocator:
    if isinstance(by, tuple):
        return by
    return by_css(by)


class FindElementsMixin:
    """Adding `element` and `elements` methods to class"""

    def element(self, by: CssSelectorOrBy) -> Element:
        """Search for single child element"""
        by = _css_to_by(by)
        return Element(SingleElementLocator(by, self))

    def elements(self, by: CssSelectorOrBy) -> ElementCollection:
        """Search for multiple child elements"""
        by = _css_to_by(by)
        return ElementCollection(MultipleElementLocator(by, self))


class Element(Searchable, FindElementsMixin, Screenshotable):
    """Single lazy element"""

    @screenshot_on_fail
    def __wait_for_condition(self, condition, timeout, exception_cls):
        exception = exception_cls(f"Condition {condition.__name__} is not reached in {timeout} seconds for {self}")
        wait_for(self, condition, timeout, exception)

    def assure(self, condition, timeout=5):
        """Make sure that element matches condition or raises :class:`TimeoutError`"""
        self.__wait_for_condition(condition, timeout, TimeoutError)

    def should(self, condition, timeout=5):
        """Make sure that element matches condition or raises :class:`AssertionError`"""
        self.__wait_for_condition(condition, timeout, AssertionError)

    should_be = should

    def _element_is_dead(self):
        try:
            _ = self.__cached__.location
            return False
        except WebDriverException:
            return True

    def get_actual(self) -> WebElement:
        """Get element, check if it's cached or already dead"""
        if self.browser.get_actual() is None:
            raise NoBrowserException
        if (self.__cached__ is None) or self._element_is_dead():
            self.__cached__ = self._search()
        return self.__cached__

    @_should_exist
    def click(self):
        """Click web element"""
        self.get_actual().click()

    @_should_exist
    def double_click(self):
        """Make double click on the element"""
        actions = ActionChains(self.browser.get_actual())
        actions.double_click(self.get_actual())

    @property
    @_should_exist
    def location(self):
        """Get current element location"""
        return self.get_actual().location

    @_should_exist
    def hover(self):
        """Hover over element"""
        actions = ActionChains(self.browser.get_actual())
        actions.move_to_element(self.get_actual()).perform()

    @_should_exist
    def right_click(self):
        """Open context menu"""
        actions = ActionChains(self.browser.get_actual())
        actions.context_click(self.get_actual()).perform()

    @property
    @_should_exist
    def text(self) -> str:
        """Get element text"""
        return self.get_actual().text

    @text.setter
    @_should_exist
    def text(self, value: str):
        """Set element text (if possible)"""
        self.clear()
        self.get_actual().send_keys(value)

    @property
    @_should_exist
    def value(self):
        """Get element @value attribute"""
        return self.get_attribute("value")

    @property
    @screenshot_on_fail
    def visible(self):
        """Check if element is visible"""
        if not self.exists:
            return False
        return self.get_actual().is_displayed()

    @property
    def hidden(self):
        """Check if element is hidden"""
        return not self.visible

    @property
    @screenshot_on_fail
    def exists(self):
        """Check if element exists in dom"""
        try:
            self.get_actual()
            return True
        except NoSuchElementException:
            return False

    @property
    @_should_exist
    def selected(self):
        """Return element selected state"""
        return self.get_actual().is_selected()

    @property
    @_should_exist
    def tag_name(self):
        """Get element tag name"""
        return self.get_actual().tag_name

    @_should_exist
    def get_attribute(self, name: str) -> str:
        """Get WebElement attribute value"""
        return str(self.get_actual().get_attribute(name))

    @_should_exist
    def clear(self):
        """Clear input field"""
        self.get_actual().clear()

    # TODO: re-enable if become useful
    # @_should_exit
    # def get_property(self, name):
    #     """Gets the given property of the element"""
    #     self.get_actual().get_property(name)

    @property
    @_should_exist
    def enabled(self):
        """Return element enabled state"""
        _enabled = self.get_actual().is_enabled()
        _disabled = (
                _not_noney(self.get_attribute("disabled"))
                or
                _not_noney(self.get_attribute("aria-disabled"))
        )
        return _enabled and not _disabled

    @property
    def disabled(self):
        """Return element disabled state"""
        return not self.enabled

    @property
    @_should_exist
    def size(self):
        """Element size"""
        return self.get_actual().size

    # TODO: re-enable if become useful
    # @_should_exit
    # def value_of_css_property(self, name):
    #     """Returns value css property"""
    #     return self.get_actual().value_of_css_property(name)

    @property
    @_should_exist
    def parent(self):
        """Return parent element of given one"""
        return self.element(by_xpath("./.."))

    def ancestors(self, filter_condition: ElementCondition = None):
        """Return parent element chain from parent element to `html`"""
        ancestors = self.elements(by_xpath("./ancestor::*"))
        if filter_condition is None:
            return ancestors
        return ancestors.filter(filter_condition)

    def neighbours(self, filter_condition: ElementCondition = None):
        """Get all elements on current level filtered by condition if given"""
        elements = self.parent.elements(by_xpath("./*"))
        if filter_condition is None:
            return elements
        return elements.filter(filter_condition)

    def capture_screenshot(self) -> bytes:
        """Capture single element screenshot"""
        try:
            return self.get_actual().screenshot_as_png
        except WebDriverException:  # if get_actual has failed
            return self.browser.capture_screenshot()

    def __repr__(self):
        return f"Element by: {repr(self._locator)}"

    def __getattr__(self, item) -> str:
        """Return value of element attribute with given name as string"""
        return self.get_attribute(item)


ElementCondition = Callable[[Element], bool]


class ElementCollection(Searchable, FindElementsMixin, Sequence):  # pylint: disable=inherit-non-class
    """Collection of lazy elements"""

    _locator: MultipleElementLocator

    def __init__(self, locator: MultipleElementLocator):  # all hail type hints
        super().__init__(locator)

    def __getitem__(self, index: Union[int, slice]) -> Union[Element, ElementCollection]:
        if isinstance(index, slice):
            return ElementCollection(SlicedElementLocator(self._locator, index))
        length = len(self)
        if 0 <= index < length:
            return Element(IndexElementLocator(self._locator, index))
        raise IndexError(f"Collection index {index} is out of range (0..{length})")

    def __len__(self) -> int:
        return len(self.get_actual())

    def __repr__(self):
        return f"Element Collection by: {repr(self._locator)}"

    def filter(self, condition: Callable[[Element], bool]):
        """Filter list of collection elements filtered by condition applied to `Element`"""
        return ElementCollection(FilteredCollectionLocator(self, condition))

    def find(self, condition: Callable[[Element], bool]):
        """Find single collection element by condition"""
        return Element(FindElementLocator(self, condition))

    def assure_all(self, condition, timeout=5, exception=TimeoutError):
        """Assure condition matches for all elements"""
        end_time = time.time() + timeout
        full_length = len(self)
        matching = None
        while time.time() < end_time:
            matching = self.filter(condition)
            if len(matching) == full_length:
                return
        raise exception(f"{full_length - len(matching)} elements are not matching condition")


def _not_noney(val):
    return str(val).lower() != 'none'
